from collections import defaultdict
from dataclasses import dataclass
from logging import getLogger
from typing import Callable

import pixeltable as pxt
import torch
from pixeltable.catalog import Table
from torch.utils.data import Dataset, DataLoader

from goldener.clusterize import GoldClusterizer
from goldener.describe import GoldDescriptor
from goldener.pxt_utils import (
    get_expr_from_column_name,
    set_value_to_idx_rows,
    GoldPxtTorchDataset,
    pxt_torch_dataset_collate_fn,
)
from goldener.select import GoldSelector
from goldener.utils import (
    check_sampling_size,
    check_all_same_type,
    get_sampling_count_from_size,
    split_sampling_among_chunks,
)
from goldener.vectorize import GoldVectorizer


logger = getLogger(__name__)


@dataclass
class GoldSet:
    """Configuration for a gold set used for splitting.

    Attributes:
        name: Name of the gold set (None if the set is for not selected data).
        size: Size (ratio or actual number) of samples to assign to this set (between 0 and 1). If a ratio, this value
        cannot be one of 0 or 1 (goal is to select a subset of the full dataset). If a number, it must be strictly
        positive and less than the total number of samples in the dataset.
    """

    name: str | None
    size: float | int

    def __post_init__(self) -> None:
        """Validate the GoldSet configuration after initialization.

        Raises:
            ValueError: If ratio is not between 0 and 1 (exclusive).
        """
        if isinstance(self.size, float) and not (0 < self.size < 1):
            raise ValueError("Size as float must be between 0 and 1.")
        elif isinstance(self.size, int) and self.size <= 0:
            raise ValueError("Size as int must be strictly positive.")


def check_sets_validity(
    sets: list[GoldSet], total: int | None = None, force_max: bool = True
) -> None:
    """Validate the sets configuration.

    This function ensures that the sum of ratios is valid and that set names are unique.

    Args:
        sets: List of GoldSet configurations to validate.
        total: Optional total number of samples (used if sizes are integers).
        force_max: Whether to enforce that the sum of ratios equals 1.0 (for float sizes) or total (for int sizes).

    Raises:
        ValueError: If the set's sizes are not valid, or if set names are not unique.
    """
    sizes = [s.size for s in sets]
    check_all_same_type(sizes)
    check_sampling_size(sum(sizes), total_size=total, force_max=force_max)

    set_names = [s.name for s in sets]
    if len(set_names) != len(set(set_names)):
        raise ValueError(f"Set names must be unique, got {set_names}")


class GoldSplitter:
    """Split a dataset into multiple sets based on features.

    The GoldSplitter leverages a GoldDescriptor to extract features from the dataset,
    a GoldVectorizer to vectorize these features, a GoldClusterizer to cluster them and a GoldSelector to select samples
    for each set based on specified ratios.

    The splitting can operate in a sequential (single-process) mode or a
    distributed mode (not implemented).

    At least 2 sets are required to split the data. Every sample
    will be associated with exactly one set using a GoldSelector. The last set is filled with any remaining elements.
    Every set is at least associated with at least 1 points (the ratio might not be fully matched for small data regimes).

    See GoldDescriptor, GoldVectorizer, and GoldSelector for more details on each component.

    Attributes:
        sets: List of GoldSet configurations defining the splits.
        selector: GoldSelector used to select samples for each set.
        descriptor: Optional GoldDescriptor used to describe the dataset.
        vectorizer: Optional GoldVectorizer used to vectorize the described dataset.
        clusterizer: Optional GoldClusterizer used to clusterize the described dataset.
        n_clusters: Number of clusters to use for clusterized selection (if clusterizer is provided).
        in_described_table: Whether to return the splitting in the described table or the selected table.
        allow_existing: Whether to allow existing tables in all components.
        drop_table: Whether to drop intermediate tables. Defaults to False.
        max_batches: Optional maximum number of batches to process in both descriptor and selector. Useful for testing on a small subset of the dataset.
    """

    def __init__(
        self,
        sets: list[GoldSet],
        selector: GoldSelector,
        descriptor: GoldDescriptor | None = None,
        vectorizer: GoldVectorizer | None = None,
        clusterizer: GoldClusterizer | None = None,
        n_clusters: int = 0,
        in_described_table: bool = False,
        allow_existing: bool = True,
        drop_table: bool = False,
        max_batches: int | None = None,
    ) -> None:
        """Initialize the GoldSplitter.

        Args:
            sets: List of GoldSet configurations defining the splits.
            selector: GoldSelector used to select samples for each set.
            descriptor: Optional GoldDescriptor used to describe the dataset.
            vectorizer: Optional GoldVectorizer used to vectorize the described dataset.
            clusterizer: Optional GoldClusterizer used to clusterize the described dataset.
            n_clusters: Number of clusters to use for clusterized selection (if clusterizer is provided).
            in_described_table: Whether to return the splitting in the described table or the selected table.
            allow_existing: Whether to allow existing tables in all components.
            drop_table: Whether to drop intermediate tables. Defaults to False.
            max_batches: Optional maximum number of batches to process for the selection.
            Useful for testing on a small subset of the dataset.
        """
        if descriptor is None and in_described_table:
            raise ValueError(
                "in_described_table is set to True, but no descriptor is provided."
            )

        if vectorizer is not None:
            if vectorizer.vectorized_key != selector.vectorized_key:
                raise ValueError(
                    f"Vectorizer vectorized_key '{vectorizer.vectorized_key}' does not match "
                    f"selector's vectorized_key '{selector.vectorized_key}'. They must be the same."
                )

            if (
                descriptor is not None
                and descriptor.description_key != vectorizer.data_key
            ):
                raise ValueError(
                    f"Descriptor description_key '{descriptor.description_key}' does not match "
                    f"vectorizer's data_key '{vectorizer.data_key}'. They must be the same."
                )

        elif (
            descriptor is not None
            and descriptor.description_key != selector.vectorized_key
        ):
            raise ValueError(
                f"Descriptor description_key '{descriptor.description_key}' does not match "
                f"selector's vectorized_key '{selector.vectorized_key}'. They must be the same."
            )

        if clusterizer is not None:
            if n_clusters <= 1:
                logger.info(
                    f"Clusterizer is provided but n_clusters is set to {n_clusters}. "
                    f"Clusterizer will not be effective. Please set n_clusters to a value greater than 1 to enable clustering."
                )

            if clusterizer.vectorized_key != selector.vectorized_key:
                raise ValueError(
                    f"Clusterizer vectorized_key '{clusterizer.vectorized_key}' does not match "
                    f"selector's vectorized_key '{selector.vectorized_key}'. They must be the same."
                )

            clusterizer.include_vectorized_in_table = True
            logger.info(
                "Clusterizer's include_vectorized_in_table is set to True to keep the "
                "vectorized features in the table for the selection step. "
                "This is necessary to perform cluster-wise selection. "
            )

        self.sets = sets
        self.descriptor = descriptor
        self.vectorizer = vectorizer
        self.clusterizer = clusterizer
        self.n_clusters = n_clusters
        self.selector = selector
        self.in_described_table = in_described_table
        self.drop_table = drop_table
        self._max_batches = max_batches
        self.allow_existing = allow_existing
        self.max_batches = max_batches

    @property
    def max_batches(self) -> int | None:
        """Get the maximum number of batches to process."""
        return self._max_batches

    @max_batches.setter
    def max_batches(self, value: int | None) -> None:
        """Set the maximum number of batches to process during splitting."""
        self._max_batches = value
        if self.descriptor is not None:
            self.descriptor.max_batches = value
        else:
            if self.vectorizer is not None:
                self.vectorizer.max_batches = value
            else:
                self.selector.max_batches = value

    @property
    def allow_existing(self) -> bool:
        """Get whether existing tables are allowed in all components."""
        return self._allow_existing

    @allow_existing.setter
    def allow_existing(self, value: bool) -> None:
        """Set whether existing tables are allowed in all components."""
        self._allow_existing = value
        if self.descriptor is not None:
            self.descriptor.allow_existing = value
        if self.vectorizer is not None:
            self.vectorizer.allow_existing = value
        self.selector.allow_existing = value

    @property
    def sets(self) -> list[GoldSet]:
        """Get the current sets configuration for the splitter."""
        return self._sets

    @sets.setter
    def sets(self, sets: list[GoldSet]) -> None:
        """Set the sets configuration for the splitter.

        Args:
            sets: New list of GoldSet configurations defining the splits.
        """
        if len(sets) < 2:
            raise ValueError("Splitting data requires at least two sets.")

        size_types = [type(s.size) for s in sets]

        if len(set(size_types)) > 1:
            raise ValueError(
                f"All set sizes must be of the same type, got types {size_types}."
            )

        check_sets_validity(
            sets, force_max=True if issubclass(size_types[0], float) else False
        )
        self._sets = sets

    @property
    def clustering_enable(self):
        return self.clusterizer is not None and self.n_clusters > 1

    @staticmethod
    def get_split_indices(
        split_data: Table | Dataset,
        selection_key: str,
        idx_key: str = "idx",
        batch_size: int = 1024,
        num_workers: int = 0,
        collate_fn: Callable | None = None,
    ) -> dict[str, set[int]]:
        """Get the indices of samples for each set from a split table.

        Args:
            split_table: PixelTable Table containing the split information.
            selection_key: Column name in the table indicating the set assignment.
            idx_key: Column name in the table indicating the sample indices.
            batch_size: Batch size to use when processing the input as dataset.
            num_workers: Number of workers to use when processing the input as dataset.
            collate_fn: Collate function to use when processing the input as dataset. It should
                return a dictionary with at least the keys specified by `idx_key` and `selection_key`.

        Returns:
            A dictionary mapping each set name to a set of sample indices.
        """

        set_indices: dict[str, set[int]] = defaultdict(set)

        if isinstance(split_data, Dataset):
            dataloader = DataLoader(
                split_data,
                batch_size=batch_size,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )

            for batch in dataloader:
                batch_indices = batch[idx_key]
                batch_selections = batch[selection_key]
                for idx_value, set_name in zip(batch_indices, batch_selections):
                    if set_name is None:
                        continue
                    set_indices[set_name].add(
                        int(idx_value.item())
                        if isinstance(idx_value, torch.Tensor)
                        else idx_value
                    )
        else:
            selection_col = get_expr_from_column_name(split_data, selection_key)
            for row in split_data.select(selection_col).distinct().collect():
                set_name = row[selection_key]
                if set_name is None:
                    continue

                set_indices[set_name] = GoldSelector.get_selection_indices(
                    split_data,
                    selection_key=selection_key,
                    value=set_name,
                    idx_key=idx_key,
                )

        return set_indices

    def split_in_dataset(
        self,
        to_split: Dataset | Table,
    ) -> GoldPxtTorchDataset:
        """Split the dataset into multiple sets in a PixelTable table.

        The dataset is first described using the gold descriptor (extracts features), and then samples are selected
        for each set based on the specified ratios after vectorization.

        At least 2 sets are required to split the data. Every sample
        will be associated with exactly one set using a GoldSelector. The last set is filled with any remaining elements.
        Every set is at least associated with at least 1 points (the ratio might not be fully matched for small data regimes).

        This method is idempotent (i.e. failure proof), meaning that if it is called
        multiple times on the same dataset or table, it will not duplicate or recompute the splitting decisions
        already present in the PixelTable table.

        Args:
            to_split: Dataset or Table to be split. If a Dataset is provided, each item should be a
            dictionary with at least the key specified by descriptor `data_key` after applying the collate_fn.
            If the collate_fn is None, the dataset is expected to directly provide such batches. If a Table is provided,
            it should contain both 'idx' and `data_key` column for the descriptor.

        Returns:
            A GoldPxtTorchDataset dataset containing at least the set assignation in the `selection_key` column. Then
            the other columns are either from the described table (if `in_described_table` is True)
            or from the selected table (if `in_described_table` is False).
        """
        split_table = self.split_in_table(to_split)

        split_dataset = GoldPxtTorchDataset(split_table, keep_cache=True)

        self._drop_tables(drop_all=True)

        return split_dataset

    def split_in_table(self, to_split: Dataset | Table) -> Table:
        """Split the dataset into multiple sets in a PixelTable table.

        The dataset is first described using the gold descriptor (extracts features), and then samples are selected
        for each set based on the specified ratios after vectorization.

        At least 2 sets are required to split the data. Every sample
        will be associated with exactly one set using a GoldSelector. The last set is filled with any remaining elements.
        Every set is at least associated with at least 1 points (the ratio might not be fully matched for small data regimes).

        This method is idempotent (i.e. failure proof), meaning that if it is called
        multiple times on the same dataset or table, it will not duplicate or recompute the splitting decisions
        already present in the PixelTable table.

        See describe_in_table, vectorize_in_table, and select_in_table for more details on each step.

        Args:
            to_split: Dataset or Table to be split. If a Dataset is provided, each item should be a
            dictionary with at least the key specified by descriptor `data_key` after applying the collate_fn.
            If the collate_fn is None, the dataset is expected to directly provide such batches. If a Table is provided,
            it should contain both 'idx' and `data_key` column for the descriptor.

        Returns:
            A PixelTable Table containing at least the set assignation in the `selection_key` column. Then
            the other columns are either from the described table (if `in_described_table` is True)
            or from the selected table (if `in_described_table` is False).

        Raises:
            ValueError: If any set results in zero samples due to its ratio, if class_key is not found,
            or if class stratification results in zero samples for any class in a set.
        """
        description = (
            self.descriptor.describe_in_table(to_split)
            if self.descriptor is not None
            else to_split
        )

        vectorized = (
            self.vectorizer.vectorize_in_table(description)
            if self.vectorizer is not None
            else description
        )

        clusterized = None
        if self.clustering_enable:
            assert self.clusterizer is not None and self.n_clusters > 1
            clusterized = self.clusterizer.cluster_in_table(vectorized, self.n_clusters)

        if isinstance(vectorized, Table):
            sample_count = vectorized.select(vectorized.idx).distinct().count()
        elif hasattr(vectorized, "__len__"):
            sample_count = len(vectorized)
        else:
            sample_count = None

        if sample_count is not None:
            check_sets_validity(self._sets, total=sample_count, force_max=True)

        # select data for all sets
        for idx_set, gold_set in enumerate(self._sets):
            logger.info(
                f"Selecting samples for set '{gold_set.name}' with size {gold_set.size}."
            )
            set_count = get_sampling_count_from_size(
                sampling_size=gold_set.size, total_size=sample_count
            )
            # the first sets are selected based on the specified ratios,
            # while the last one gathers all the remaining samples,
            # so we don't need to select it explicitly
            if idx_set < len(self._sets) - 1:
                if clusterized is not None:
                    selected_table = self._clusterized_selection(
                        clusterized, set_count, gold_set
                    )
                else:
                    selected_table = self.selector.select_in_table(
                        vectorized, set_count, value=gold_set.name
                    )

                if sample_count is None:
                    sample_count = (
                        selected_table.select(selected_table.idx).distinct().count()
                    )
                    check_sets_validity(self._sets, total=sample_count, force_max=True)
            else:
                # The last set is gathering all the remaining samples
                selection_col = get_expr_from_column_name(
                    selected_table, self.selector.selection_key
                )
                already_in_set_count = selected_table.where(
                    selection_col == gold_set.name
                ).count()
                not_yet_selected = selected_table.where(selection_col == None)  # noqa: E711
                not_yet_selected_count = not_yet_selected.count()
                if not_yet_selected_count == 0 and already_in_set_count == 0:
                    raise ValueError(
                        f"Not enough data to split among {len(self._sets)} sets."
                    )

                if (already_in_set_count + not_yet_selected_count) != set_count:
                    logger.warning(
                        f"For the set '{gold_set.name}', the expected count was {set_count}, "
                        f"but got {already_in_set_count} already selected and {not_yet_selected_count} not yet selected samples. "
                        f"This might be due to rounding issues."
                    )

                not_yet_selected.update(  # noqa: E711
                    {self.selector.selection_key: gold_set.name}
                )

        # Depending on the `in_described_table` flag, move the selection column to the described table
        # or keep it in the selected table.
        split_table = selected_table
        if self.in_described_table:
            if not isinstance(description, Table):
                raise ValueError(
                    "in_described_table is set to True, but description is not a PixelTable Table."
                )
            for set_cfg in self._sets:
                set_name = set_cfg.name
                set_indices = self.selector.get_selection_indices(
                    selected_table,
                    selection_key=self.selector.selection_key,
                    value=set_name,
                )
                if self.selector.selection_key not in description.columns():
                    description.add_column(
                        if_exists="error",
                        **{self.selector.selection_key: pxt.String},
                    )
                set_value_to_idx_rows(
                    description,
                    col_expr=get_expr_from_column_name(
                        description, self.selector.selection_key
                    ),
                    idx_expr=description.idx,
                    indices=set_indices,
                    value=set_name,
                )
            split_table = description

        self._drop_tables(drop_all=False)

        return split_table

    def _clusterized_selection(
        self,
        clusterized: Table,
        set_count: int,
        gold_set: GoldSet,
    ) -> Table:
        assert self.clusterizer is not None
        # the table for each cluster are processed sequentially, and sent as GoldPxtTorchDataset
        # to the selector, so the collate_fn of the selector is temporarily updated to `pxt_torch_dataset_collate_fn,
        # and reset to the original collate_fn after the cluster-wise selection is done
        selection_collate_fn = self.selector.collate_fn
        logger.info(
            "Clusterized selection is enabled. Temporarily setting selector's collate_fn "
            "to pxt_torch_dataset_collate_fn to process cluster-wise selection. "
            "It will be reset to the original collate_fn after the cluster-wise selection is done."
        )
        self.selector.collate_fn = pxt_torch_dataset_collate_fn

        try:
            # define all the queries, select count and indices for each cluster
            cluster_col = get_expr_from_column_name(
                clusterized, self.clusterizer.cluster_key
            )
            if clusterized.where(cluster_col == None).count():  # noqa: E711
                raise ValueError(
                    f"Clusterizer output column '{self.clusterizer.cluster_key}' contains null values."
                )

            available_count = clusterized.select(clusterized.idx).distinct().count()
            still_to_select_count = set_count

            # run selection for each cluster and gather the selected indices
            # the selected table will be filled incrementally
            already_selected: set[int] = set()
            selected_table = None  # ensure it exists in the scope after the loop, even if no cluster is selected
            for cluster_idx in range(self.n_clusters):
                # define the number of sample to select for the cluster
                # previous clusters will have reduced the available count for the next clusters,
                cluster_indices = self.clusterizer.get_cluster_indices(
                    table=clusterized,
                    cluster_key=self.clusterizer.cluster_key,
                    cluster_idx=cluster_idx,
                    idx_key="idx",
                )
                cluster_indices = cluster_indices - already_selected
                if len(cluster_indices) == 0:
                    logger.warning(
                        f"Cluster {cluster_idx} has no available samples for selection. Skipping this cluster."
                    )
                    continue
                cluster_select_count = split_sampling_among_chunks(
                    still_to_select_count,
                    [len(cluster_indices), available_count - len(cluster_indices)],
                )[0]
                if cluster_select_count == 0:
                    logger.warning(
                        f"Cluster {cluster_idx} has no samples assigned for selection. Skipping this cluster."
                    )
                    continue

                selected_table = self.selector.select_in_table(
                    GoldPxtTorchDataset(
                        clusterized.where(
                            (cluster_col == cluster_idx)
                            & (clusterized.idx.isin(cluster_indices))
                        )
                    ),
                    cluster_select_count,
                    value=gold_set.name,
                )

                # update the elements allowing to compute the selection specs for the next clusters
                already_selected = self.selector.get_selection_indices(
                    selected_table,
                    selection_key=self.selector.selection_key,
                    value=gold_set.name,
                )
                selected_count = len(already_selected)
                still_to_select_count -= selected_count
                available_count -= selected_count

        finally:
            # restore the original collate_fn of the selector after cluster-wise selection is done
            self.selector.collate_fn = selection_collate_fn

        if selected_table is None:
            raise RuntimeError(
                f"no selection table was created during clusterized selection for set '{gold_set.name}'."
            )

        return selected_table

    def _drop_tables(self, drop_all: bool = True) -> None:
        """Drop all intermediate tables created during the splitting process.

        Args:
            drop_all: Whether to drop all tables including the selected table.
            The selected table is either the selector's table or the descriptor's table
            depending on the `in_described_table` flag.
        """
        if self.drop_table:
            table_names = [
                self.selector.table_path,
                *([self.descriptor.table_path] if self.descriptor is not None else []),
                *([self.vectorizer.table_path] if self.vectorizer is not None else []),
                *(
                    [self.clusterizer.table_path]
                    if self.clusterizer is not None
                    else []
                ),
            ]

            to_not_drop = (
                self.descriptor.table_path
                if self.in_described_table and self.descriptor is not None
                else self.selector.table_path
            )

            for table in table_names:
                if table != to_not_drop or drop_all:
                    pxt.drop_table(table, if_not_exists="ignore")



import math
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Callable, Any

import numpy as np
import pixeltable as pxt
from pixeltable import Error
from pixeltable.catalog import Table

import torch
from sklearn.base import ClusterMixin
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm

from goldener.pxt_utils import (
    set_value_to_idx_rows,
    GoldPxtTorchDataset,
    get_expr_from_column_name,
    get_valid_table,
    make_batch_ready_for_table,
    check_pxt_table_has_primary_key,
    get_sample_row_from_idx,
)
from goldener.reduce import GoldReducer
from goldener.torch_utils import get_dataset_sample_dict
from goldener.utils import (
    filter_batch_from_indices,
)


logger = getLogger(__name__)


class GoldClusteringTool(ABC):
    """Run clustering on input vectors."""

    @abstractmethod
    def fit(self, x: torch.Tensor, n_clusters: int) -> torch.Tensor:
        """Fit the clustering tool from a given set of vectors into n clusters.

        Args:
            x: Input vectors to select from.
            n_clusters: Number of clusters to form.

        Returns: The cluster assignments for each input vector as a 1D tensor of cluster indices.
        """
        pass

    @abstractmethod
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """Assign the input vectors to the clusters.

        Args:
            x: Input vectors to transform.

        Returns: The cluster assignments for each input vector as a 1D tensor of cluster indices.
        """
        pass


class GoldRandomClusteringTool(GoldClusteringTool):
    """Chunk data randomly into clusters of almost equal size."""

    def __init__(self, random_state: int | None = None) -> None:
        self.random_state = random_state

    def fit(self, x: torch.Tensor, n_clusters: int) -> torch.Tensor:
        """Randomly assign each input vector to clusters of roughly the same size.

        Args:
            x: Input vectors to select from.
            n_clusters: Number of clusters to form.

        Returns: The cluster assignments for each input vector as a 1D tensor of cluster indices.
        """
        total = len(x)

        if n_clusters > total:
            raise ValueError(
                f"Number of clusters ({n_clusters}) cannot be greater than "
                f"the number of samples ({total})."
            )

        cluster_assignment = [i % n_clusters for i in range(total)]
        np.random.default_rng(self.random_state).shuffle(cluster_assignment)

        return torch.tensor(
            cluster_assignment,
        )

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("Random clustering tool does not support predict.")


class GoldSKLearnClusteringTool(GoldClusteringTool):
    """Chunk data randomly into clusters of almost equal size."""

    def __init__(self, tool: ClusterMixin) -> None:
        if not hasattr(tool, "predict"):
            raise ValueError(
                f"The clustering tool {tool} must provide a predict method."
            )

        if not hasattr(tool, "n_clusters"):
            raise NotImplementedError(
                f"The clustering tool {tool} must allow specifying `n_clusters`."
            )

        self.tool = tool

    def fit(self, x: torch.Tensor, n_clusters: int) -> torch.Tensor:
        """Fit the scikit-learn clustering tool on the input vectors and return cluster assignments.

        Args:
            x: Input vectors to select from.
            n_clusters: Number of clusters to form.

        Returns: The cluster assignments for each input vector as a 1D tensor of cluster indices.
        """
        self.tool.n_clusters = n_clusters
        x_np = x.detach().cpu().numpy()
        labels = self.tool.fit(x_np).labels_
        return torch.from_numpy(labels)

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """Assign the input vectors to clusters using the fitted scikit-learn clustering tool.

        Args:
            x: Input vectors to assign to clusters.

        Returns:
            A 1D tensor of cluster indices predicted for each input vector.
        """
        x_np = x.detach().cpu().numpy()
        return torch.from_numpy(self.tool.predict(x_np))


class GoldClusterizer:
    """Cluster data points from vectorized samples.

    The GoldClusterizer processes a dataset or PixelTable table to perform clustering using a
    clustering algorithm on already vectorized representations. The clustering results are stored
    in a local PixelTable table (specified by `table_path`) so that the clustering process is idempotent:
    calling the same operation multiple times will not duplicate or recompute clustering that are already
    present in the table.

    If the dataset is too big to fit into memory or the clustering algorithm is too
    computationally expensive, the clustering can be performed in chunks.
    All torch tensors will be converted to numpy arrays before saving.

    The clustering can operate in a sequential (single-process) mode or a
    distributed mode (not implemented). The table schema is created/validated automatically and will include
    minimal indexing columns (`idx`, `idx_vector`) required to link clustered samples back to
    their originating data points.

    Attributes:
        table_path: Path to the PixelTable table where clustering results will be stored locally.
        clustering_tool: The GoldClusteringTool implementing the clustering algorithm.
        reducer: Optional GoldReducer instance for dimensionality reduction before clustering.
        chunk: Optional chunk size for processing data in chunks to reduce memory consumption.
        collate_fn: Optional function to collate dataset samples into batches composed of
            dictionaries with at least the key specified by `vectorized_key` returning a PyTorch Tensor.
            If None, the dataset is expected to directly provide such batches.
        vectorized_key: Key in the batch dictionary that contains the vectorized data for the clustering. Default is "vectorized".
        include_vectorized_in_table: Whether to keep the vectorized data in the cluster table.
            It is only applied if the cluster table is created from a Table (it is forced anyway for Dataset). Default is False.
        cluster_key: Column name to store the clustering value in the PixelTable table. Default is "cluster".
        class_key: Optional key for class-based stratified clustering.
        to_keep_schema: Optional dictionary defining additional columns to keep from the original dataset/table
            into the cluster table. The keys are the column names and the values are the PixelTable types.
        min_pxt_insert_size: Minimum number of rows to accumulate before inserting into PixelTable. Default is 100.
        batch_size: Batch size used when iterating over the data.
        num_workers: Number of workers for the PyTorch DataLoader during iteration on data.
        allow_existing: If False, an error will be raised when the table already exists. Default is True.
        distribute: Whether to use distributed processing for clustering and table population. Not implemented yet. Default is False.
        drop_table: Whether to drop the cluster table after creating the dataset with clustering results. It is only applied
            when using `cluster_in_dataset`. Default is False.
        max_batches: Optional maximum number of batches to process. Useful for testing on a small subset of the dataset.
        random_state: Optional random state for reproducibility during chunk assignment.
    """

    _MINIMAL_SCHEMA: dict[str, type] = {
        "idx": pxt.Required[pxt.Int],
        "idx_vector": pxt.Required[pxt.Int],
    }

    def __init__(
        self,
        table_path: str,
        clustering_tool: GoldClusteringTool,
        reducer: GoldReducer | None = None,
        chunk: int | None = None,
        collate_fn: Callable | None = None,
        vectorized_key: str = "vectorized",
        include_vectorized_in_table: bool = False,
        cluster_key: str = "cluster",
        class_key: str | None = None,
        to_keep_schema: dict[str, type] | None = None,
        min_pxt_insert_size: int = 100,
        batch_size: int = 1,
        num_workers: int = 0,
        allow_existing: bool = True,
        distribute: bool = False,
        drop_table: bool = False,
        max_batches: int | None = None,
        random_state: int | None = None,
    ) -> None:
        """Initialize the GoldClusterizer with configuration parameters.

        Attributes:
            table_path: Path to the PixelTable table where clustering results will be stored locally.
            clustering_tool: The GoldClusteringTool implementing the clustering algorithm.
            reducer: Optional GoldReducer instance for dimensionality reduction before clustering.
            chunk: Optional chunk size for processing data in chunks to reduce memory consumption.
            collate_fn: Optional function to collate dataset samples into batches composed of
                dictionaries with at least the key specified by `vectorized_key` returning a PyTorch Tensor.
                If None, the dataset is expected to directly provide such batches.
            vectorized_key: Key in the batch dictionary that contains the vectorized data for the clustering. Default is "vectorized".
            include_vectorized_in_table: Whether to keep the vectorized data in the cluster table. Defaults to False.
                    It is only applied if the cluster table is created from a Table (it is forced anyway for Dataset).
            cluster_key: Column name to store the clustering value in the PixelTable table. Default is "cluster".
            class_key: Optional key for class-based stratified clustering.
            to_keep_schema: Optional dictionary defining additional columns to keep from the original dataset/table
                into the cluster table. The keys are the column names and the values are the PixelTable types.
            min_pxt_insert_size: Minimum number of rows to accumulate before inserting into PixelTable. Default is 100.
            batch_size: Batch size used when iterating over the data.
            num_workers: Number of workers for the PyTorch DataLoader during iteration on data.
            allow_existing: If False, an error will be raised when the table already exists. Default is True.
            distribute: Whether to use distributed processing for clustering and table population. Not implemented yet. Default is False.
            drop_table: Whether to drop the cluster table after creating the dataset with clustering results. It is only applied
                when using `cluster_in_dataset`. Default is False.
            max_batches: Optional maximum number of batches to process. Useful for testing on a small subset of the dataset.
            random_state: Optional random state for reproducibility during chunk assignment.
        """
        self.table_path = table_path
        self.clustering_tool = clustering_tool
        self.reducer = reducer
        if chunk is not None and chunk <= 0:
            raise ValueError("chunk must be a positive integer or None.")
        self.chunk = chunk
        self.collate_fn = collate_fn
        self.vectorized_key = vectorized_key
        self.include_vectorized_in_table = include_vectorized_in_table
        self.cluster_key = cluster_key
        self.class_key = class_key
        self.to_keep_schema = to_keep_schema
        self.min_pxt_insert_size = min_pxt_insert_size
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.allow_existing = allow_existing
        self.distribute = distribute
        self.drop_table = drop_table
        self.max_batches = max_batches
        self.random_state = random_state

    def cluster_in_dataset(
        self, cluster_from: Dataset | Table, n_clusters: int
    ) -> GoldPxtTorchDataset:
        """Cluster the data and return results as a GoldPxtTorchDataset.

        The clustering process applies a clustering algorithm on already vectorized
        representations of the data points and stores results in a PixelTable table specified by `table_path`.
        When the chunk attribute is set, the clustering is performed in chunks to reduce memory consumption.
        If a reducer is provided, the vectors are reduced in dimension before applying the clustering.

        This is a convenience wrapper that runs `cluster_in_table` to populate
        (or resume populating) the PixelTable table, then wraps the table into a
        `GoldPxtTorchDataset` for downstream consumption. If `drop_table` is True,
        the table will be removed after the dataset is created.

        Args:
            cluster_from: Dataset or Table to cluster. If a Dataset is provided, each item should be a
                dictionary with at least the `vectorized_key`, `idx_vector` and `idx` keys after applying the collate_fn.
                If the collate_fn is None, the dataset is expected to directly provide such batches.
                If a Table is provided, it should contain at least the `vectorized_key` and `idx` columns.
            n_clusters: Number of clusters to create.
        Returns:
            A GoldPxtTorchDataset containing at least the clustering information in the `cluster_key` key
                and `idx` (index of the sample) and `idx_vector` (index of the vector) keys as well.
        """

        cluster_table = self.cluster_in_table(cluster_from, n_clusters)

        cluster_dataset = GoldPxtTorchDataset(cluster_table, keep_cache=True)

        if self.drop_table:
            pxt.drop_table(cluster_table)

        return cluster_dataset

    def cluster_in_table(self, cluster_from: Dataset | Table, n_clusters: int) -> Table:
        """Cluster the data and store results in a PixelTable table.

        The clustering process applies a clustering algorithm on already vectorized
        representations of the data points and stores results in a PixelTable table specified by `table_path`.
        When the chunk attribute is set, the clustering is performed in chunks to reduce memory consumption.
        If a reducer is provided, the vectors are reduced in dimension before applying the clustering.

        This method is idempotent (i.e., failure proof), meaning that if it is called
        multiple times on the same dataset or table, it will restart the clustering process
        based on the vectors already present in the PixelTable table.

        Args:
            cluster_from: Dataset or Table to select from. If a Dataset is provided, each item should be a
                dictionary with at least the `vectorized_key` and `idx` keys after applying the collate_fn.
                If the collate_fn is None, the dataset is expected to directly provide such batches.
                If a Table is provided, it should contain at least the `vectorized_key`, `idx` and `idx_vector` columns.
            n_clusters: Number of clusters to create.

        Returns:
            A PixelTable Table containing at least the clustering information in the `cluster_key` column
                and `idx` (index of the sample) and `idx_vector` (index of the vector) columns as well.

        Raises:
            ValueError: If n_clusters is less than or equal to 1.
        """
        if n_clusters <= 1:
            raise ValueError(
                "cluster_in_table requires n_clusters to be greater than 1."
            )

        logger.info(f"Loading the existing clustering table from {self.table_path}")
        try:
            old_cluster_table = pxt.get_table(
                self.table_path,
                if_not_exists="ignore",
            )
        except Error:
            logger.info(f"No existing clustering table from {self.table_path}")
            old_cluster_table = None

        # clustering table is expected to have a primary key allowing to idempotent updates
        if old_cluster_table is not None:
            check_pxt_table_has_primary_key(old_cluster_table, {"idx_vector"})

        if not self.allow_existing and old_cluster_table is not None:
            raise ValueError(
                f"Table at path {self.table_path} already exists and "
                "allow_existing is set to False."
            )

        if isinstance(cluster_from, Table):
            cluster_table = self._cluster_table_from_table(
                cluster_from=cluster_from,
                old_cluster_table=old_cluster_table,
            )
        else:
            cluster_table = self._cluster_table_from_dataset(
                cluster_from, old_cluster_table
            )
            cluster_from = cluster_table

        assert isinstance(cluster_from, Table)

        # define the number of element to sample
        total_size = cluster_from.select(cluster_from.idx_vector).distinct().count()

        if len(self.get_cluster_indices(cluster_table, self.cluster_key)) == total_size:
            logger.info(f"Cluster table {self.table_path} already fully clustered")
            return cluster_table
        elif self.distribute:
            self._distributed_cluster(cluster_from, cluster_table, n_clusters)
        else:
            self._sequential_cluster(cluster_from, cluster_table, n_clusters)

        logger.info(
            f"Cluster table {self.table_path} successfully clustered in {n_clusters} clusters."
        )

        return cluster_table

    def _cluster_table_from_table(
        self, cluster_from: Table, old_cluster_table: Table | None
    ) -> Table:
        """Create or validate the cluster table schema from a PixelTable table.

        This private method sets up the table structure with necessary columns for tracking
        clustering status and ensures all rows from the source table are represented.

        Args:
            cluster_from: The source PixelTable table to cluster.
            old_cluster_table: Existing cluster table if resuming, or None.

        Returns:
            The cluster table with proper schema and initial rows.
        """
        minimal_schema = self._MINIMAL_SCHEMA.copy()

        if self.to_keep_schema is not None:
            minimal_schema |= self.to_keep_schema

        if self.class_key is not None:
            minimal_schema[self.class_key] = pxt.String

        cluster_table = get_valid_table(
            table=old_cluster_table
            if old_cluster_table is not None
            else self.table_path,
            minimal_schema=minimal_schema,
            primary_key="idx_vector",
        )

        if self.vectorized_key not in cluster_from.columns():
            raise ValueError(
                f"Table at path {self.table_path} does not contain "
                f"the required column {self.vectorized_key}."
            )

        cluster_table_cols = cluster_table.columns()
        if self.cluster_key not in cluster_table_cols:
            cluster_table.add_column(if_exists="error", **{self.cluster_key: pxt.Int})

        if (
            self.include_vectorized_in_table
            and self.vectorized_key not in cluster_table_cols
        ):
            idx_vectors = [
                row["idx_vector"] for row in cluster_from.sample(1).collect()
            ]
            if not idx_vectors:
                raise ValueError(
                    f"Source table at {self.table_path} is empty, cannot sample vectorized data for schema inference."
                )
            sample = get_sample_row_from_idx(
                cluster_from,
                idx_key="idx_vector",
                # make sure to take an existing vector
                idx=idx_vectors[0],
                collate_fn=self.collate_fn,
                expected_keys=[self.vectorized_key],
            )

            vectorized_value = sample[self.vectorized_key]
            cluster_table.add_column(
                **{
                    self.vectorized_key: pxt.Array[  # type: ignore[misc]
                        vectorized_value.shape, pxt.Float
                    ]
                }
            )

        if cluster_table.count() > 0:
            to_cluster_indices = set(
                [
                    row["idx_vector"]
                    for row in cluster_from.select(cluster_from.idx_vector)
                    .distinct()
                    .collect()
                ]
            )
            already_in_cluster_table = set(
                [
                    row["idx_vector"]
                    for row in cluster_table.select(cluster_table.idx_vector)
                    .distinct()
                    .collect()
                ]
            )
            still_to_cluster = to_cluster_indices.difference(already_in_cluster_table)
            if not still_to_cluster:
                logger.info(
                    f"The cluster table is already initialized in {self.table_path}"
                )
                return cluster_table

        col_list = ["idx", self.vectorized_key]
        if self.to_keep_schema is not None:
            col_list.extend(list(self.to_keep_schema.keys()))

        if self.cluster_key in cluster_from.columns():
            col_list.append(self.cluster_key)

        if self.class_key is not None and self.class_key in cluster_from.columns():
            col_list.append(self.class_key)

        self._add_rows_to_cluster_table_from_dataset(
            cluster_from=GoldPxtTorchDataset(
                cluster_from.select(
                    *[
                        get_expr_from_column_name(cluster_from, col)
                        for col in col_list + ["idx_vector"]
                    ]
                ),
                keep_cache=False,
            ),
            cluster_table=cluster_table,
            include_vectorized=self.include_vectorized_in_table,
        )

        return cluster_table

    def _cluster_table_from_dataset(
        self, cluster_from: Dataset, old_cluster_table: Table | None
    ) -> Table:
        """Create or validate the cluster table schema from a PyTorch Dataset.

        This private method sets up the table structure with necessary columns
        and validate that the vectorized column is present in the dataset.

        Args:
            cluster_from: The source PyTorch Dataset to select from.
            old_cluster_table: Existing clustering table if resuming, or None.

        Returns:
            The clustering table with proper schema.
        """
        minimal_schema = self._MINIMAL_SCHEMA.copy()
        if self.to_keep_schema is not None:
            minimal_schema |= self.to_keep_schema

        if self.class_key is not None:
            minimal_schema[self.class_key] = pxt.String

        cluster_table = get_valid_table(
            table=old_cluster_table
            if old_cluster_table is not None
            else self.table_path,
            minimal_schema=minimal_schema,
            primary_key="idx_vector",
        )
        cluster_table_cols = cluster_table.columns()
        if self.vectorized_key not in cluster_table_cols:
            sample = get_dataset_sample_dict(
                cluster_from,
                collate_fn=self.collate_fn,
                expected=[self.vectorized_key],
            )

            vectorized_value = sample[self.vectorized_key].detach().cpu().numpy()
            cluster_table.add_column(
                **{
                    self.vectorized_key: pxt.Array[  # type: ignore[misc]
                        vectorized_value.shape, pxt.Float
                    ]
                }
            )

        if self.cluster_key not in cluster_table_cols:
            cluster_table.add_column(if_exists="error", **{self.cluster_key: pxt.Int})

        self._add_rows_to_cluster_table_from_dataset(
            cluster_from, cluster_table, include_vectorized=True
        )

        return cluster_table

    def _add_rows_to_cluster_table_from_dataset(
        self,
        cluster_from: Dataset,
        cluster_table: Table,
        include_vectorized: bool = False,
    ) -> None:
        """Add rows from the source dataset to the cluster table.

        This private method iterates through the dataset in batches and populates the
        cluster table with vectorized data and metadata, skipping already processed samples.

        Args:
            cluster_from: The source PyTorch Dataset.
            cluster_table: The clustering table to populate.
            include_vectorized: Whether to include the vectorized data in the cluster table.
        """
        dataloader = DataLoader(
            cluster_from,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            collate_fn=self.collate_fn,
        )

        already_in_clustering = set(
            [
                row["idx_vector"]
                for row in cluster_table.select(cluster_table.idx_vector).collect()
            ]
        )
        not_empty = (
            len(already_in_clustering) > 0
        )  # allow to filter out already described samples

        ready_to_insert: list[dict[str, Any]] = []

        for batch_idx, batch in tqdm(
            enumerate(dataloader),
            desc="Initializing rows for the clustering table",
            total=(
                None if hasattr(cluster_from, "__len__") is False else len(dataloader)
            ),
        ):
            # Stop if we've processed enough batches
            if self.max_batches is not None and batch_idx >= self.max_batches:
                break

            if "idx_vector" not in batch:
                starts = batch_idx * self.batch_size
                batch["idx_vector"] = [
                    starts + idx for idx in range(len(batch[self.vectorized_key]))
                ]

            if "idx" not in batch:
                batch["idx"] = batch["idx_vector"]

            # Keep only not yet included samples in the batch
            if not_empty:
                batch = filter_batch_from_indices(
                    batch,
                    already_in_clustering,
                    index_key="idx_vector",
                )

                if len(batch) == 0:
                    continue  # all samples already described

            if self.cluster_key not in batch:
                batch[self.cluster_key] = [
                    None for _ in range(len(batch[self.vectorized_key]))
                ]

            already_in_clustering.update(
                [
                    idx.item() if isinstance(idx, torch.Tensor) else idx
                    for idx in batch["idx_vector"]
                ]
            )
            to_insert_keys = ["idx", self.cluster_key]
            if self.to_keep_schema is not None:
                to_insert_keys.extend(list(self.to_keep_schema.keys()))
            if self.class_key is not None and self.class_key in batch:
                to_insert_keys.append(self.class_key)
            if include_vectorized:
                to_insert_keys.append(self.vectorized_key)

            batch_as_list = make_batch_ready_for_table(
                batch,
                to_insert_keys,
                "idx_vector",
            )

            ready_to_insert.extend(batch_as_list)
            if len(ready_to_insert) >= self.min_pxt_insert_size:
                cluster_table.insert(ready_to_insert)
                ready_to_insert = []

        if ready_to_insert:
            cluster_table.insert(ready_to_insert)

    @staticmethod
    def get_cluster_indices(
        table: Table,
        cluster_key: str,
        cluster_idx: int | None = None,
        class_key: str | None = None,
        class_value: str | None = None,
        idx_key: str = "idx_vector",
    ) -> set[int]:
        """Get the indices of samples clustered in a given cluster.

        Args:
            table: PixelTable table to query.
            cluster_key: Column name used to store the clustering values.
            cluster_idx: Optional cluster index to filter samples by cluster.
            If None, all clustered samples are returned.
            class_key: Optional column name used to filter samples by class.
            class_value: Optional class value to filter samples by class.
            idx_key: Column name used to get sample indices.
        """
        idx_col = get_expr_from_column_name(table, idx_key)
        cluster_col = get_expr_from_column_name(table, cluster_key)
        query = (
            cluster_col != None  # noqa: E711
            if cluster_idx is None
            else cluster_col == cluster_idx
        )
        if class_value is not None and class_key is not None:
            class_col = get_expr_from_column_name(table, class_key)
            query = query & (class_col == class_value)
        else:
            if class_key is not None or class_value is not None:
                raise ValueError("class_key and class_value must be set together.")

        return set(
            [
                row[idx_key]
                for row in table.where(query)  # noqa: E712
                .select(idx_col)
                .distinct()
                .collect()
            ]
        )

    def _sequential_cluster(
        self,
        cluster_from: Table,
        cluster_table: Table,
        n_clusters: int,
    ) -> None:
        """Run sequential (single-process) clustering process.

        This private method handles class-stratified clustering if a class_key is configured,
        otherwise performs clustering on the full dataset. It delegates the actual clustering
        to _class_cluster.

        Args:
            cluster_from: The source table with vectorized data.
            cluster_table: The table to store clustering results.
            n_clusters: Number of clusters.
        """
        if self.class_key is not None:
            class_col = get_expr_from_column_name(cluster_table, self.class_key)
            class_values = [
                distinct_item[self.class_key]
                for distinct_item in cluster_table.select(class_col)
                .distinct()
                .collect()
            ]

            for class_idx, class_value in enumerate(class_values):
                already_clustered = len(
                    self.get_cluster_indices(
                        table=cluster_table,
                        cluster_key=self.cluster_key,
                        class_key=self.class_key,
                        class_value=class_value,
                    )
                )
                class_still_to_cluster_count = (
                    cluster_table.where(
                        class_col == class_value  # noqa: E712
                    )
                    .select(cluster_table.idx_vector)
                    .distinct()
                    .count()
                )

                class_still_to_cluster_count = (
                    class_still_to_cluster_count - already_clustered
                )
                if class_still_to_cluster_count == 0:
                    logger.info(
                        f"Class '{class_value}' already fully clustered, skipping."
                    )
                    continue
                elif class_still_to_cluster_count < 0:
                    raise ValueError(
                        "The size of the cluster table has decreased since the 1st clustering computation"
                    )

                self._cluster_class(
                    cluster_from,
                    cluster_table,
                    n_clusters,
                    class_value=class_value,
                )

        else:
            already_clustered = len(
                self.get_cluster_indices(
                    table=cluster_table,
                    cluster_key=self.cluster_key,
                )
            )
            sample_count = cluster_table.count()
            still_to_cluster_count = sample_count - already_clustered
            logger.info(
                f"Clustering {still_to_cluster_count} samples in {n_clusters} clusters."
            )
            self._cluster_class(
                cluster_from,
                cluster_table,
                n_clusters,
            )

    def _cluster_class(
        self,
        cluster_from: Table,
        cluster_table: Table,
        n_clusters: int,
        class_value: str | None = None,
    ) -> None:
        """Perform clustering for a specific class or all data.

        This private method implements chunked clustering.
        It processes data in chunks to manage memory, applies optional dimensionality reduction,
        and updates the cluster table with clustered samples.

        Args:
            cluster_from: The source table with vectorized data.
            cluster_table: The table to store clustering results.
            n_clusters: Number of clusters.
            class_value: Optional class value to filter samples by class.

        Raises:
            ValueError: If an unexpected empty chunk is encountered during clustering.
        """
        cluster_col = get_expr_from_column_name(cluster_table, self.cluster_key)
        vectorized_col = get_expr_from_column_name(cluster_from, self.vectorized_key)

        if class_value is not None:
            assert self.class_key is not None
            class_col = get_expr_from_column_name(cluster_table, self.class_key)
            available_query = (cluster_col == None) & (class_col == class_value)  # noqa: E712 E711
        else:
            available_query = cluster_col == None  # noqa: E711

        to_cluster = cluster_table.where(available_query)
        to_cluster_vector_indices = [
            row["idx_vector"]
            for row in to_cluster.select(cluster_table.idx_vector).collect()
        ]
        available_for_clustering = len(to_cluster_vector_indices)
        if available_for_clustering == 0:
            logger.info(
                f"No samples to cluster for class '{class_value}'."
                if class_value is not None
                else "No samples to cluster."
            )
            return

        if self.chunk is None or self.chunk >= len(to_cluster_vector_indices):
            chunk_assignment = [
                to_cluster_vector_indices,
            ]
        else:
            chunk_count = math.ceil(available_for_clustering / self.chunk)
            random_assignment = (
                GoldRandomClusteringTool(random_state=self.random_state)
                .fit(
                    torch.empty(
                        available_for_clustering, 1
                    ),  # dummy input for random clustering
                    chunk_count,
                )
                .tolist()
            )
            chunk_assignment = [
                [
                    vector_idx
                    for vect_pos, vector_idx in enumerate(to_cluster_vector_indices)
                    if random_assignment[vect_pos] == chunk_idx
                ]
                for chunk_idx in range(chunk_count)
            ]

        for chunk_indices in chunk_assignment:
            to_cluster_from = cluster_from.where(
                cluster_from.idx_vector.isin(chunk_indices)
            ).select(vectorized_col, cluster_from.idx_vector)

            if to_cluster_from.count() == 0:
                raise ValueError("Unexpected empty chunk for clustering.")

            to_cluster_for_chunk = [
                (
                    torch.from_numpy(sample[self.vectorized_key]),
                    torch.tensor(sample["idx_vector"]).unsqueeze(0),
                )
                for sample in to_cluster_from.collect()
            ]
            vectors_list, indices_list = map(list, zip(*to_cluster_for_chunk))
            vectors = torch.stack(vectors_list, dim=0)
            indices = torch.cat(indices_list, dim=0)

            if self.reducer is not None:
                vectors = self.reducer.fit_transform(vectors)

            cluster_assignment = self.clustering_tool.fit(vectors, n_clusters)

            # update table with cluster information
            for cluster_idx in set(cluster_assignment.tolist()):
                indices_in_cluster = indices[cluster_assignment == cluster_idx].tolist()
                set_value_to_idx_rows(
                    table=cluster_table,
                    col_expr=cluster_col,
                    idx_expr=cluster_table.idx_vector,
                    indices=set(indices_in_cluster),
                    value=cluster_idx,
                )

    def _distributed_cluster(
        self,
        cluster_from: Table,
        cluster_table: Table,
        n_clusters: int,
    ) -> None:
        """Run distributed clustering process (not implemented).

        Args:
            cluster_from: The source table with vectorized data.
            cluster_table: The table to store the clustering results.
            n_clusters: Number of clusters.

        Raises:
            NotImplementedError: Always raised as distributed mode is not yet implemented.
        """
        raise NotImplementedError("Distributed clustering is not implemented yet.")



import pytest
import torch
import pixeltable as pxt

from torch.utils.data import Dataset

from goldener.clusterize import (
    GoldRandomClusteringTool,
    GoldClusterizer,
    GoldSKLearnClusteringTool,
)
from goldener.reduce import GoldReducer
from goldener.pxt_utils import GoldPxtTorchDataset


class DummyDataset(Dataset):
    def __init__(self, samples):
        self._samples = samples

    def __len__(self):
        return len(self._samples)

    def __getitem__(self, idx):
        return (
            self._samples[idx].copy()
            if isinstance(self._samples[idx], dict)
            else self._samples[idx]
        )


class TestGoldRandomClusteringTool:
    def test_fit(self):
        n_clusters = 2
        total = 10
        tool = GoldRandomClusteringTool(42)
        clusters = tool.fit(torch.randn(total, 3), n_clusters)
        assert clusters.shape == (total,)
        assert set(clusters.tolist()) == {0, 1}
        assert len([c for c in clusters if c == 0]) == total / n_clusters
        assert not torch.equal(
            clusters, torch.tensor([i % n_clusters for i in range(total)])
        )

    def test_predict(self):
        with pytest.raises(NotImplementedError):
            GoldRandomClusteringTool(42).predict(torch.randn(10, 3))

    def test_with_nclusters_greater_than_total(self):
        n_clusters = 15
        total = 10
        tool = GoldRandomClusteringTool(42)
        with pytest.raises(
            ValueError, match="cannot be greater than the number of samples"
        ):
            tool.fit(torch.randn(total, 3), n_clusters)


class TestGoldSKLearnClusteringTool:
    def test_init_raises_when_tool_has_no_predict(self):
        class NoPredict:
            def __init__(self) -> None:
                self.n_clusters = 2

        with pytest.raises(ValueError, match="must provide a predict method"):
            GoldSKLearnClusteringTool(NoPredict())

    def test_init_raises_when_tool_has_no_n_clusters(self):
        class NoNClusters:
            def predict(self, x):
                return x

        with pytest.raises(
            NotImplementedError, match="must allow specifying `n_clusters`"
        ):
            GoldSKLearnClusteringTool(NoNClusters())

    def test_fit_returns_tensor_with_expected_shape_and_labels(self):
        from sklearn.cluster import KMeans

        n_samples = 20
        n_clusters = 3
        x = torch.randn(n_samples, 4)

        tool = GoldSKLearnClusteringTool(KMeans(n_clusters=n_clusters, random_state=0))
        labels = tool.fit(x, n_clusters)

        assert isinstance(labels, torch.Tensor)
        assert labels.shape == (n_samples,)
        # all labels in expected range
        assert set(labels.tolist()).issubset(set(range(n_clusters)))

    def test_predict_matches_sklearn_predict_shape(self):
        from sklearn.cluster import KMeans

        n_samples = 10
        n_clusters = 2
        x = torch.randn(n_samples, 5)

        base = KMeans(n_clusters=n_clusters, random_state=0)
        tool = GoldSKLearnClusteringTool(base)

        # fit on half of the samples, then predict on all
        tool.fit(x[:5], n_clusters)
        preds = tool.predict(x)

        assert isinstance(preds, torch.Tensor)
        assert preds.shape == (n_samples,)
        # ensure we actually used sklearn under the hood by checking range
        assert set(preds.tolist()).issubset(set(range(n_clusters)))


class TestGoldClusterizer:
    def setup_method(self):
        pxt.drop_dir("unit_test", force=True)
        pxt.create_dir("unit_test", if_exists="ignore")

    def teardown_method(self):
        pxt.drop_dir("unit_test", force=True)

    def _make_src_table(self, path: str, n: int = 10, with_class: bool = False):
        source_rows = []
        for i in range(n):
            row = {
                "idx": i,
                "idx_vector": i,
                "vectorized": torch.rand(4).numpy(),
            }
            if with_class:
                row["label"] = str(i % 2)
            source_rows.append(row)
        return pxt.create_table(path, source=source_rows, if_exists="replace_force")

    def test_cluster_table_creation_from_table(self):
        src_path = "unit_test.src_cluster_table_input"
        cluster_path = "unit_test.test_cluster_from_table"

        src_table = self._make_src_table(src_path, n=5)

        clusterizer = GoldClusterizer(
            table_path=cluster_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=False,
        )

        cluster_table = clusterizer._cluster_table_from_table(
            cluster_from=src_table,
            old_cluster_table=None,
        )

        assert set(cluster_table.columns()) >= {
            "idx",
            "idx_vector",
            clusterizer.cluster_key,
        }
        row_indices = [
            row["idx_vector"]
            for row in cluster_table.select(cluster_table.idx_vector).collect()
        ]
        assert set(row_indices) == set(range(5))

    def test_cluster_table_from_table_when_missing_vectorized(self):
        src_path = "unit_test.src_cluster_table_invalid"
        src_table = pxt.create_table(
            src_path,
            source=[{"idx": 0, "idx_vector": 0, "notvec": [1, 2, 3]}],
            if_exists="replace_force",
        )

        clusterizer = GoldClusterizer(
            table_path="unit_test.test_cluster",
            clustering_tool=GoldRandomClusteringTool(),
        )

        with pytest.raises(ValueError, match="does not contain the required column"):
            clusterizer._cluster_table_from_table(
                cluster_from=src_table,
                old_cluster_table=None,
            )

    def test_cluster_table_from_dataset(self):
        table_path = "unit_test.test_cluster_initialize"

        sample = {"vectorized": torch.rand(4), "idx": 0}
        dataset = DummyDataset([sample, sample])

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=False,
        )

        cluster_table = clusterizer._cluster_table_from_dataset(
            cluster_from=dataset,
            old_cluster_table=None,
        )

        assert set(cluster_table.columns()) >= {
            clusterizer.vectorized_key,
            clusterizer.cluster_key,
            "idx",
            "idx_vector",
        }
        row_indices = [
            row["idx_vector"]
            for row in cluster_table.select(cluster_table.idx_vector).collect()
        ]
        assert set(row_indices) == {0, 1}

    def test_cluster_in_table_from_dataset(self):
        table_path = "unit_test.test_cluster_from_dataset"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(4), "idx": idx} for idx in range(20)]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=5,
        )

        cluster_table = clusterizer.cluster_in_table(dataset, n_clusters=4)

        assert cluster_table.count() == 20
        assert (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
            == 20
        )
        distinct_clusters = [
            row[clusterizer.cluster_key]
            for row in cluster_table.select(cluster_table[clusterizer.cluster_key])
            .distinct()
            .collect()
        ]
        assert set(distinct_clusters).issubset(set(range(4)))

    def test_cluster_in_table_from_table(self):
        src_path = "unit_test.src_cluster_table"
        cluster_path = "unit_test.test_cluster_from_table_full"

        src_table = self._make_src_table(src_path, n=20)

        clusterizer = GoldClusterizer(
            table_path=cluster_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
        )

        cluster_table = clusterizer.cluster_in_table(src_table, n_clusters=4)

        assert cluster_table.count() == 20
        assert (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
            == 20
        )
        distinct_clusters = [
            row[clusterizer.cluster_key]
            for row in cluster_table.select(cluster_table[clusterizer.cluster_key])
            .distinct()
            .collect()
        ]
        assert set(distinct_clusters).issubset(set(range(4)))

    def test_cluster_in_table_from_table_with_vectorized_included(self):
        src_path = "unit_test.src_cluster_table"
        cluster_path = "unit_test.test_cluster_from_table_full"

        src_table = self._make_src_table(src_path, n=20)

        clusterizer = GoldClusterizer(
            table_path=cluster_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            include_vectorized_in_table=True,
        )

        cluster_table = clusterizer.cluster_in_table(src_table, n_clusters=4)

        assert cluster_table.count() == 20
        assert (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
            == 20
        )
        distinct_clusters = [
            row[clusterizer.cluster_key]
            for row in cluster_table.select(cluster_table[clusterizer.cluster_key])
            .distinct()
            .collect()
        ]
        assert set(distinct_clusters).issubset(set(range(4)))

    def test_cluster_in_table_with_invalid_n_clusters(self):
        table_path = "unit_test.test_cluster_invalid_n"
        dataset = DummyDataset(
            [{"vectorized": torch.rand(4), "idx": idx} for idx in range(10)]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
        )

        with pytest.raises(ValueError, match="n_clusters to be greater than 1"):
            clusterizer.cluster_in_table(dataset, n_clusters=1)

    def test_cluster_in_table_with_chunk(self):
        table_path = "unit_test.test_cluster_chunk"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(4), "idx": idx} for idx in range(50)]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=10,
            chunk=10,
        )

        cluster_table = clusterizer.cluster_in_table(dataset, n_clusters=5)

        assert cluster_table.count() == 50
        assert (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
            == 50
        )
        distinct_clusters = [
            row[clusterizer.cluster_key]
            for row in cluster_table.select(cluster_table[clusterizer.cluster_key])
            .distinct()
            .collect()
        ]
        assert set(distinct_clusters).issubset(set(range(5)))

    def test_cluster_in_table_with_chunk_larger_than_total(self):
        table_path = "unit_test.test_cluster_chunk_large"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(4), "idx": idx} for idx in range(15)]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=5,
            chunk=100,  # larger than total number of vectors (15)
        )

        cluster_table = clusterizer.cluster_in_table(dataset, n_clusters=3)

        assert cluster_table.count() == 15
        clustered = (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
        )
        assert clustered == 15

        distinct_clusters = [
            row[clusterizer.cluster_key]
            for row in cluster_table.select(cluster_table[clusterizer.cluster_key])
            .distinct()
            .collect()
        ]
        assert set(distinct_clusters).issubset(set(range(3)))

    def test_cluster_in_table_with_reducer(self):
        table_path = "unit_test.test_cluster_reducer"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(8), "idx": idx} for idx in range(30)]
        )

        from sklearn.decomposition import PCA

        reducer = GoldReducer(PCA(n_components=4))

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=10,
            reducer=reducer,
        )

        cluster_table = clusterizer.cluster_in_table(dataset, n_clusters=3)

        assert cluster_table.count() == 30
        assert (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
            == 30
        )
        distinct_clusters = [
            row[clusterizer.cluster_key]
            for row in cluster_table.select(cluster_table[clusterizer.cluster_key])
            .distinct()
            .collect()
        ]
        assert set(distinct_clusters).issubset(set(range(3)))

    def test_cluster_in_dataset_and_drop_table(self):
        table_path = "unit_test.test_cluster_dataset_drop"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(4), "idx": idx} for idx in range(10)]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=5,
            drop_table=True,
        )

        clustered_dataset = clusterizer.cluster_in_dataset(dataset, n_clusters=2)

        assert isinstance(clustered_dataset, GoldPxtTorchDataset)

        first_item = next(iter(clustered_dataset))
        assert isinstance(first_item, dict)
        assert "idx" in first_item

        with pytest.raises(pxt.Error):
            pxt.get_table(table_path)

    def test_cluster_in_table_with_existing_table_and_disallow(self):
        src_path = "unit_test.src_cluster_table_existing"
        cluster_path = "unit_test.test_cluster_existing"

        src_table = self._make_src_table(src_path, n=5)

        pxt.create_table(
            cluster_path,
            source=[{"idx": 0, "idx_vector": 0, "cluster": 0}],
            if_exists="replace_force",
        )

        clusterizer = GoldClusterizer(
            table_path=cluster_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=False,
        )

        with pytest.raises(
            ValueError, match="does not have all specified columns as primary keys"
        ):
            clusterizer.cluster_in_table(src_table, n_clusters=2)

    def test_cluster_in_table_with_restart_allowed(self):
        table_path = "unit_test.test_cluster_restart_allowed"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(4), "idx": idx} for idx in range(30)]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=False,
            batch_size=5,
        )

        cluster_table_1 = clusterizer.cluster_in_table(dataset, n_clusters=3)

        assert cluster_table_1.count() == 30
        clustered_count_1 = (
            cluster_table_1.where(
                cluster_table_1[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table_1.idx)
            .distinct()
            .count()
        )
        assert clustered_count_1 == 30

        clusterizer_restart = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=5,
        )

        cluster_table_2 = clusterizer_restart.cluster_in_table(dataset, n_clusters=3)

        assert cluster_table_2.count() == 30
        clustered_count_2 = (
            cluster_table_2.where(
                cluster_table_2[clusterizer_restart.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table_2.idx)
            .distinct()
            .count()
        )
        assert clustered_count_2 == 30
        assert clustered_count_2 == clustered_count_1

    def test_cluster_in_table_with_class_key(self):
        table_path = "unit_test.test_cluster_with_class"

        dataset = DummyDataset(
            [
                {
                    "vectorized": torch.rand(4),
                    "idx": idx,
                    "label": str(idx % 2),
                }
                for idx in range(40)
            ]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=8,
            class_key="label",
        )

        cluster_table = clusterizer.cluster_in_table(dataset, n_clusters=4)

        assert cluster_table.count() == 40
        assert (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
            == 40
        )

        label_col = cluster_table[clusterizer.class_key]
        total_class0 = (
            cluster_table.where(label_col == "0")
            .select(cluster_table.idx)
            .distinct()
            .count()
        )
        total_class1 = (
            cluster_table.where(label_col == "1")
            .select(cluster_table.idx)
            .distinct()
            .count()
        )
        assert total_class0 == 20
        assert total_class1 == 20

        all_indices = clusterizer.get_cluster_indices(
            table=cluster_table,
            cluster_key=clusterizer.cluster_key,
        )
        assert all_indices == set(range(40))

        class0_indices = clusterizer.get_cluster_indices(
            table=cluster_table,
            cluster_key=clusterizer.cluster_key,
            class_key=clusterizer.class_key,
            class_value="0",
        )
        class1_indices = clusterizer.get_cluster_indices(
            table=cluster_table,
            cluster_key=clusterizer.cluster_key,
            class_key=clusterizer.class_key,
            class_value="1",
        )

        assert class0_indices.union(class1_indices) == set(range(40))
        assert class0_indices.isdisjoint(class1_indices)

    def test_cluster_in_table_with_max_batches(self):
        table_path = "unit_test.test_cluster_max_batches"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(4), "idx": idx} for idx in range(50)]
        )

        clusterizer = GoldClusterizer(
            table_path=table_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            batch_size=10,
            max_batches=2,
        )

        cluster_table = clusterizer.cluster_in_table(dataset, n_clusters=3)

        clustered = (
            cluster_table.where(
                cluster_table[clusterizer.cluster_key] != None  # noqa: E711
            )
            .select(cluster_table.idx)
            .distinct()
            .count()
        )
        assert clustered == 20

    def test_get_cluster_indices(self):
        src_path = "unit_test.test_cluster_indices_explicit_src"
        cluster_path = "unit_test.test_cluster_indices_explicit"

        src_table = self._make_src_table(src_path, n=6, with_class=True)

        clusterizer = GoldClusterizer(
            table_path=cluster_path,
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            allow_existing=True,
            class_key="label",
        )

        cluster_table = clusterizer._cluster_table_from_table(
            cluster_from=src_table,
            old_cluster_table=None,
        )

        cluster_table.where(cluster_table.idx_vector.isin([0, 1, 2])).update(
            {clusterizer.cluster_key: 0}
        )
        cluster_table.where(cluster_table.idx_vector.isin([3, 4, 5])).update(
            {clusterizer.cluster_key: 1}
        )

        all_indices = clusterizer.get_cluster_indices(
            table=cluster_table,
            cluster_key=clusterizer.cluster_key,
        )
        assert all_indices == {0, 1, 2, 3, 4, 5}

        class0_indices = clusterizer.get_cluster_indices(
            table=cluster_table,
            cluster_key=clusterizer.cluster_key,
            class_key=clusterizer.class_key,
            class_value="0",
        )
        class1_indices = clusterizer.get_cluster_indices(
            table=cluster_table,
            cluster_key=clusterizer.cluster_key,
            class_key=clusterizer.class_key,
            class_value="1",
        )

        assert class0_indices.union(class1_indices) == {0, 1, 2, 3, 4, 5}
        assert class0_indices.isdisjoint(class1_indices)

        with pytest.raises(ValueError, match="must be set together"):
            clusterizer.get_cluster_indices(
                table=cluster_table,
                cluster_key=clusterizer.cluster_key,
                class_key=clusterizer.class_key,
            )
